#| code-fold: true
from micrograd2023.engine import Value
from micrograd2023.nn import Layer
import random
def test_unit_layer():
    
    # random seed for reproducibility
    random.seed(4899)
    # input dimension
    Ninput = 5
    Nout = 3
    # input vector
    xs = [Value(random.random()) for _ in range(Ninput)]
    # call a layer
    nl = Layer(nin=Ninput, nout=Nout)

    # number of parameters should be (Ninput + 1 bias)xNout = (5+1)x3 = 18
    num_paras =  (Ninput+1)*Nout
    assert len(nl.parameters()) == num_paras

    ys = nl(xs)
    # test out put dimension
    assert len(ys) == Nout

    # manual calculation of the outputs
    ym = [sum(xi*wi for xi, wi in zip(xs, nl.neurons[idx].parameters())) for idx in range(Nout)]
    # manual Relu()
    ym = [y if y.data >0 else Value(0) for y in ym]

    # test ys (output of neuron) vs. ym (manual calculation)
    tol = 1e-6
    for ysi, ymi in zip(ys, ym):
        assert abs(ysi.data - ymi.data) < tol



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_engine.ipynb.

# %% auto 0
__all__ = ['Value']

# %% ../nbs/01_engine.ipynb 3
import math
# import torch

# %% ../nbs/01_engine.ipynb 4
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *
from .utils import *

# %% ../nbs/01_engine.ipynb 7
class Value:
    """ stores a single scalar value and its gradient """

    def __init__(self, data, _children=(), _op=''):
        self.data = data
        self.grad = 0
        # internal variables used for autograd graph construction
        self._backward = lambda: None
        self._prev = set(_children)
        self._op = _op # the op that produced this node, for graphviz / debugging / etc

    def __add__(self, other):
        other = other if isinstance(other, Value) else Value(other)
        out = Value(self.data + other.data, (self, other), '+')

        def _backward():
            self.grad += out.grad
            other.grad += out.grad
        out._backward = _backward

        return out

    def __mul__(self, other):
        other = other if isinstance(other, Value) else Value(other)
        out = Value(self.data * other.data, (self, other), '*')

        def _backward():
            self.grad += other.data * out.grad
            other.grad += self.data * out.grad
        out._backward = _backward

        return out

    def __pow__(self, other):
        assert isinstance(other, (int, float)), "only supporting int/float powers for now"
        out = Value(self.data**other, (self,), f'**{other}')

        def _backward():
            self.grad += (other * self.data**(other-1)) * out.grad
        out._backward = _backward

        return out

    def relu(self):
        out = Value(0 if self.data < 0 else self.data, (self,), 'ReLU')

        def _backward():
            self.grad += (out.data > 0) * out.grad
        out._backward = _backward

        return out

    # HPD - adding tanh func from Andrej Lecture
    def tanh(self):
        x = self.data
        t = (math.exp(2*x) - 1)/(math.exp(2*x) + 1)
        out = Value(t, (self, ), 'tanh')
        
        def _backward():
            self.grad += (1 - t**2) * out.grad
        out._backward = _backward
        
        return out
    
    # HPD - adding exponential function
    def exp(self):
        x = self.data
        t = math.exp(x)
        out = Value(t, (self, ), 'exp')
        
        def _backward():
            self.grad += t * out.grad
        out._backward = _backward
            
        return out
    
    # HPD - adding natural logarithm based 'e' function
    def log(self):
        x = self.data
        t = math.log(x)
        out = Value(t, (self, ), 'log')
        
        def _backward():
            self.grad += (1/self.data) * out.grad
        out._backward = _backward
            
        return out

    # HPD - adding sin(x) function
    def sin(self):
        x = self.data
        t = math.sin(x)
        out = Value(t, (self, ), 'sin')
        
        def _backward():
            self.grad += (math.cos(self.data)) * out.grad
        out._backward = _backward
            
        return out
    
    # HPD - adding cos(x) function
    def cos(self):
        x = self.data
        t = math.cos(x)
        out = Value(t, (self, ), 'cos')
        
        def _backward():
            self.grad += (-math.sin(self.data)) * out.grad
        out._backward = _backward
            
        return out
    
    # HPD - adding zero_grad function
    def zero_grad(self):
        self.grad = 0  # Set gradient to zero

    def backward(self):

        # topological order all of the children in the graph
        topo = []
        visited = set()
        def build_topo(v):
            if v not in visited:
                visited.add(v)
                for child in v._prev:
                    build_topo(child)
                topo.append(v)
        build_topo(self)

        # go one variable at a time and apply the chain rule to get its gradient
        self.grad = 1
        for v in reversed(topo):
            v._backward()

    def __neg__(self): # -self
        return self * -1

    def __radd__(self, other): # other + self
        return self + other

    def __sub__(self, other): # self - other
        return self + (-other)

    def __rsub__(self, other): # other - self
        return other + (-self)

    def __rmul__(self, other): # other * self
        return self * other

    def __truediv__(self, other): # self / other
        return self * other**-1

    def __rtruediv__(self, other): # other / self
        return other * self**-1

    def __repr__(self):
        return f"Value(data={self.data}, grad={self.grad})"



#| code-fold: true
from micrograd2023.engine import Value
from micrograd2023.nn import MLP
import random
def test_unit_mlp():
    # random seed for reproducibility
    random.seed(4876)
    # input dimension
    Ninput = 3
    Nhid1, Nhid2 = 5, 4 
    Nout = 2
    # input vector
    xs = [Value(random.random()) for _ in range(Ninput)]
    # call a MLP
    nnw = MLP(nin=Ninput, nouts = [Nhid1, Nhid2, Nout])

    # test number of layer
    assert len(nnw.layers) == len([Nhid1, Nhid2, Nout])
    num_layers = len(nnw.layers)

    # # number of parameters should be (Ninput + 1 bias)*Nhid1 + (Nhid1+1 bias)*Nhid2 + (Nhid2 + 1bias)*Nout
    num_paras =  (Ninput + 1)*Nhid1 + (Nhid1+1)*Nhid2 + (Nhid2 + 1)*Nout
    assert len(nnw.parameters()) == num_paras

    ys = nnw(xs)
    # # test out put dimension
    assert len(ys) == Nout

    # # manual calculation of the outputs
    xh = xs
    for idx in range(num_layers):
        ym = nnw.layers[idx](xh)
        xh = ym
    # no Relu() for the output layer
    # ym = [y if y.data >0 else Value(0) for y in ym]

    # test ys (output of neuron) vs. ym (manual calculation)
    tol=1.e-6
    for ysi, ymi in zip(ys, ym):
        assert abs(ysi.data - ymi.data) < tol 



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_utils.ipynb.

# %% auto 0
__all__ = ['trace', 'draw_dot', 'derivatives_matrix', 'fisher_info_matrix', 'cramer_rao_bound', 'CRLB_pipeline']

# %% ../nbs/03_utils.ipynb 5
from graphviz import Digraph

# %% ../nbs/03_utils.ipynb 8
def trace(root):
    nodes, edges = set(), set()
    def build(v):
        if v not in nodes:
            nodes.add(v)
            for child in v._prev:
                edges.add((child, v))
                build(child)
    build(root)
    return nodes, edges

def draw_dot(root, format='svg', rankdir='LR'):
    """
    format: png | svg | ...
    rankdir: TB (top to bottom graph) | LR (left to right)
    """
    assert rankdir in ['LR', 'TB']
    nodes, edges = trace(root)
    dot = Digraph(format=format, graph_attr={'rankdir': rankdir}) #, node_attr={'rankdir': 'TB'})
    
    for n in nodes:
        dot.node(name=str(id(n)), label = "{ data %.4f | grad %.4f }" % (n.data, n.grad), shape='record')
        if n._op:
            dot.node(name=str(id(n)) + n._op, label=n._op)
            dot.edge(str(id(n)) + n._op, str(id(n)))
    
    for n1, n2 in edges:
        dot.edge(str(id(n1)), str(id(n2)) + n2._op)
    
    return dot

# %% ../nbs/03_utils.ipynb 14
import numpy as np
import inspect

# %% ../nbs/03_utils.ipynb 15
def derivatives_matrix(func, ts: np.ndarray, params: dict, param_labels: list) -> np.ndarray:
    """calculate the derivatives matrix of a function w.r.t. its parameters at given measurements at ts"""
    # Initialize a dict to store grads for each parameter
    grads = {label: np.zeros(ts.size) for label in param_labels}
    
    for idx, t in enumerate(ts):
        # auto-diff
        y = func(t, *params.values())
        y.backward()
        
        # store grad for each parameter in the grads dictionary
        for label in param_labels:
            grads[label][idx] = params[label].grad
            params[label].zero_grad()
    
    # derivatives matrix
    Dij = np.zeros((ts.size, len(param_labels)))
    for i, label in enumerate(param_labels):
        Dij[:, i] = grads[label]
    
    return Dij

# %% ../nbs/03_utils.ipynb 16
def fisher_info_matrix(Dij: np.ndarray, sigma: float) -> np.ndarray:
    """calculate the Fisher Information Matrix (FIM) from the derivatives matrix Dij"""
    Iij = 1/(sigma**2) * Dij.T @ Dij
    return Iij

# %% ../nbs/03_utils.ipynb 17
def cramer_rao_bound(Iij: np.ndarray) -> np.ndarray:
    """calculate the Cramer-Rao Lower Bound (CRLB) from the Fisher Information Matrix (FIM)"""
    CRLB = np.linalg.inv(Iij)
    return CRLB

# %% ../nbs/03_utils.ipynb 18
def CRLB_pipeline(func, ts: np.ndarray, params: dict, param_labels: list, sigma:float, cond_num_cutoff=1e15) -> np.ndarray:
    """calculate the Cramer-Rao Lower Bound (CRLB) from a given function, measurements, parameters, and noise std"""
    Dij = derivatives_matrix(func, ts, params, param_labels)
    Iij = fisher_info_matrix(Dij, sigma)
    # condition number check
    if np.linalg.cond(Iij) < cond_num_cutoff:
        CRLB = cramer_rao_bound(Iij)
    else:
        print(f'Condition number of the Fisher Information Matrix is too high: {np.linalg.cond(Iij)}')
        CRLB = np.full(Iij.shape, np.nan)
    return CRLB



#| code-fold: true
from micrograd2023.engine import Value
from micrograd2023.nn import Neuron
import random
def test_unit_neuron():
    
    # random seed for reproducibility
    random.seed(41)
    # input dimension
    Ninput = 5

    # input vector
    xs = [Value(random.random()) for _ in range(Ninput)]

    # call a neuron
    nron = Neuron(nin=Ninput)

    # number of parameters should be 6 (5 + 1 bias)
    assert len(nron.parameters()) == (Ninput+1)

    # neuron output    
    ys = nron(xs)
    # manual calculation of the output
    ym = sum(xi*wi for xi, wi in zip(xs, nron.parameters()))
    # manual Relu()
    ym = ym if ym.data >0 else Value(0)
    print(ys, ym)

    # test ys (output of neuron) vs. ym (manual calculation)
    tol = 1e-6
    assert abs(ys.data - ym.data) < tol



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_nn.ipynb.

# %% auto 0
__all__ = ['Module', 'Neuron', 'Layer', 'MLP']

# %% ../nbs/02_nn.ipynb 3
from nbdev.showdoc import *
from fastcore.test import *
from fastcore.utils import *
from IPython import display

# %% ../nbs/02_nn.ipynb 5
import random
from .engine import Value

# %% ../nbs/02_nn.ipynb 6
from .utils import draw_dot

# %% ../nbs/02_nn.ipynb 17
class Module:
    """base module"""
    def zero_grad(self):
        for p in self.parameters():
            p.grad = 0

    def parameters(self):
        return []

# %% ../nbs/02_nn.ipynb 20
class Neuron(Module):
    """Artificial Neuron e.g. Perceptron"""
    
    def __init__(self, nin, nonlin=True):
        self.w = [Value(random.uniform(-1,1)) for _ in range(nin)]
        self.b = Value(0)
        self.nonlin = nonlin

    def __call__(self, x):
        act = sum((wi*xi for wi,xi in zip(self.w, x)), self.b)
        return act.relu() if self.nonlin else act

    def parameters(self):
        return self.w + [self.b]

    def __repr__(self):
        return f"{'ReLU' if self.nonlin else 'Linear'}Neuron({len(self.w)})"

# %% ../nbs/02_nn.ipynb 29
class Layer(Module):
    """Layer of Neurons"""
    def __init__(self, nin, nout, nonlin=True, **kwargs):
        self.neurons = [Neuron(nin, nonlin=nonlin, **kwargs) for _ in range(nout)]

    def __call__(self, x):
        out = [n(x) for n in self.neurons]
        return out[0] if len(out) == 1 else out

    def parameters(self):
        return [p for n in self.neurons for p in n.parameters()]

    def __repr__(self):
        return f"Layer of [{', '.join(str(n) for n in self.neurons)}]"

# %% ../nbs/02_nn.ipynb 38
class MLP(Module):
    """Multi-Layer Perceptron (MLP) e.g. Neural Network"""

    def __init__(self, nin, nouts, nonlin=True):
        sz = [nin] + nouts
        if nonlin:
            self.layers = [Layer(sz[i], sz[i+1], nonlin=i!=len(nouts)-1) for i in range(len(nouts))]
        else:
            self.layers = [Layer(sz[i], sz[i+1], nonlin=nonlin) for i in range(len(nouts))]
            
    def __call__(self, x):
        for layer in self.layers:
            x = layer(x)
        return x

    def parameters(self):
        return [p for layer in self.layers for p in layer.parameters()]

    def __repr__(self):
        return f"MLP of [{', '.join(str(layer) for layer in self.layers)}]"



#| code-fold: true
import torch
from micrograd2023.engine import Value
def test_integration_engine():

    x = Value(-4.0)
    z = 2 * x + 2 + x
    q = z.relu() + z * x
    h = (z * z).relu()
    y = h + q + q * x
    y.backward()
    xmg, ymg = x, y

    x = torch.Tensor([-4.0]).double()
    x.requires_grad = True
    z = 2 * x + 2 + x
    q = z.relu() + z * x
    h = (z * z).relu()
    y = h + q + q * x
    y.backward()
    xpt, ypt = x, y

    # forward pass went well
    assert ymg.data == ypt.data.item()
    # backward pass went well
    assert xmg.grad == xpt.grad.item()
    
def test_integration_engine_more():

    a = Value(-4.0)
    b = Value(2.0)
    c = a + b
    d = a * b + b**3
    c += c + 1
    c += 1 + c + (-a)
    d += d * 2 + (b + a).relu()
    d += 3 * d + (b - a).relu()
    e = c - d
    f = e**2
    g = f / 2.0
    g += 10.0 / f
    g.backward()
    amg, bmg, gmg = a, b, g

    a = torch.Tensor([-4.0]).double()
    b = torch.Tensor([2.0]).double()
    a.requires_grad = True
    b.requires_grad = True
    c = a + b
    d = a * b + b**3
    c = c + c + 1
    c = c + 1 + c + (-a)
    d = d + d * 2 + (b + a).relu()
    d = d + 3 * d + (b - a).relu()
    e = c - d
    f = e**2
    g = f / 2.0
    g = g + 10.0 / f
    g.backward()
    apt, bpt, gpt = a, b, g

    tol = 1e-6
    # forward pass went well
    assert abs(gmg.data - gpt.data.item()) < tol
    # backward pass went well
    assert abs(amg.grad - apt.grad.item()) < tol
    assert abs(bmg.grad - bpt.grad.item()) < tol



#| code-fold: true
import torch
from micrograd2023.engine import Value
from micrograd2023.nn import MLP
import random
import numpy as np
def test_integration_nn():
    """ TODO: to refactor this function and make it cleaner :) """
    # random seed for reproducibility
    random.seed(48198)
    # input dimension
    Ninput = 3
    Nhid1, Nhid2 = 5, 4 
    Nout = 2
    # input vector
    xs = [Value(random.random()) for _ in range(Ninput)]
    # call a MLP
    nnw = MLP(nin=Ninput, nouts = [Nhid1, Nhid2, Nout])
    ys = nnw(xs)
    loss = sum(ys).tanh()
    loss.backward()


    # let's have the same inputs for torch
    xst = torch.Tensor([xsi.data for xsi in xs])
    xst.requires_grad = True

    # let's use the same model for torch
    tmodel = torch.nn.Sequential(
        torch.nn.Linear(Ninput, Nhid1),
        torch.nn.ReLU(),
        torch.nn.Linear(Nhid1, Nhid2),
        torch.nn.ReLU(),
        torch.nn.Linear(Nhid2, Nout),
        # torch.nn.Tanh()
    )
    # let's fill the same parameters for the torch model
    ## 1st hidden layer
    tlayer_id = 0
    layer_id = 0
    Nrow = Nhid1
    Ncol = Ninput
    ws = []
    for idx in range(len(nnw.layers[layer_id].parameters())):
        ws.append(nnw.layers[layer_id].parameters()[idx].data)
    # reshape and remove the bias column
    wss = np.reshape(ws, (Nrow, Ncol+1))[:,:-1]
    # fill in the same weights for tmodel
    for i in range(Nrow):
            for j in range(Ncol):
                tmodel[tlayer_id].weight.data[i,j].fill_(wss[i,j])

    tmodel[tlayer_id].bias.data.fill_(0.0)

    # let's fill the same parameters for the torch model
    ## 2nd hidden layer
    tlayer_id = 2
    layer_id = 1
    Nrow = Nhid2
    Ncol = Nhid1
    ws = []
    for idx in range(len(nnw.layers[layer_id].parameters())):
        ws.append(nnw.layers[layer_id].parameters()[idx].data)
    # reshape and remove the bias column
    wss = np.reshape(ws, (Nrow, Ncol+1))[:,:-1]
    # fill in the same weights for tmodel
    for i in range(Nrow):
            for j in range(Ncol):
                tmodel[tlayer_id].weight.data[i,j].fill_(wss[i,j])

    tmodel[tlayer_id].bias.data.fill_(0.0)

    # let's fill the same parameters for the torch model
    ## 3rd hidden layer
    tlayer_id = 4
    layer_id = 2
    Nrow = Nout
    Ncol = Nhid2
    ws = []
    for idx in range(len(nnw.layers[layer_id].parameters())):
        ws.append(nnw.layers[layer_id].parameters()[idx].data)
    # reshape and remove the bias column
    wss = np.reshape(ws, (Nrow, Ncol+1))[:,:-1]
    # fill in the same weights for tmodel
    for i in range(Nrow):
            for j in range(Ncol):
                tmodel[tlayer_id].weight.data[i,j].fill_(wss[i,j])

    tmodel[tlayer_id].bias.data.fill_(0.0)

    # compute loss for torch
    yst = tmodel(xst)
    tloss = yst.sum().tanh()
    tloss.backward()

    # test forward pass
    tol = 1e-6
    assert abs(loss.data - tloss.data.item()) < tol

    # test backward pass
    for idx in range(Ninput):
        assert abs(xs[idx].grad - xst.grad[idx]) < tol



#| code-fold: true
import torch
from micrograd2023.engine import Value
def test_unit_tanh():
    a = Value(1.15)
    b = Value(5.33)

    g = a * b.tanh()
    g.backward()
    amg, bmg, gmg = a, b, g
    print(amg.grad, bmg.grad, gmg.data)

    a = torch.Tensor([1.15]).double()
    b = torch.Tensor([5.33]).double()
    a.requires_grad = True
    b.requires_grad = True

    g = a * torch.tanh(b)
    g.backward()
    apt, bpt, gpt = a, b, g
    print(apt.grad.item(), bpt.grad.item(), gpt.data.item())

    tol = 1e-4
    # test forward pass
    assert abs(gmg.data - gpt.data.item()) < tol 
    # test backward pass i.e. derivatives of the output w.r.t. parameters a and b
    assert abs(amg.grad - apt.grad.item()) < tol 
    assert abs(bmg.grad - bpt.grad.item()) < tol
